package org.openmuc.jdlms.internal.lnassociation;

import java.io.EOFException;
import java.io.IOException;
import java.net.SocketTimeoutException;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Semaphore;

import org.openmuc.jdlms.ServerConnectionInfo.Status;
import org.openmuc.jdlms.ServerConnectionListener;
import org.openmuc.jdlms.internal.APdu;
import org.openmuc.jdlms.internal.DataDirectory;
import org.openmuc.jdlms.internal.DataDirectory.DlmsLogicalDevice;
import org.openmuc.jdlms.internal.ServerConnectionData;
import org.openmuc.jdlms.internal.ServiceError;
import org.openmuc.jdlms.internal.StateError;
import org.openmuc.jdlms.internal.asn1.cosem.COSEMpdu;
import org.openmuc.jdlms.internal.asn1.cosem.COSEMpdu.Choices;
import org.openmuc.jdlms.internal.sessionlayer.ServerSessionLayer;
import org.openmuc.jdlms.internal.settings.ServerSettings;
import org.openmuc.jdlms.internal.transportlayer.ServerConnectionInformationImpl;

public class Association implements Runnable {

    private final DataDirectory directory;
    private final ServerSessionLayer sessionLayer;

    private final Semaphore semaphore;
    private final Long connectionId;
    private final ServerConnectionData connectionData;
    private ServerSettings settings;
    private ServerConnectionInformationImpl serverConnectionInformation;

    public Association(DataDirectory directory, ServerSessionLayer sessionLayer, Semaphore semaphore, Long connectionId,
            ServerSettings settings, ServerConnectionInformationImpl serverConnectionInformation) {
        this.directory = directory;
        this.sessionLayer = sessionLayer;
        this.semaphore = semaphore;
        this.connectionId = connectionId;
        this.settings = settings;
        this.serverConnectionInformation = serverConnectionInformation;

        this.connectionData = new ServerConnectionData();
        this.directory.addConnection(connectionId, this.connectionData);

    }

    @Override
    public void run() {
        AssociationMessenger associationMessenger = new AssociationMessenger(connectionData, directory, sessionLayer,
                connectionId);
        ServerConnectionListener connectionListener = settings.connectionListener;

        try {
            byte[] payload = this.sessionLayer.readNextMessage();
            Status status = Status.OPEN;

            notifyListener(connectionListener, status);

            RequestProcessorData requestProcessorData = new RequestProcessorData(this.sessionLayer.getLogicalDeviceId(),
                    connectionId, directory, connectionData);

            this.connectionData.clientId = this.sessionLayer.getClientId();
            DlmsLogicalDevice dlmsLogicalDevice = this.directory
                    .getLogicalDeviceFor(this.sessionLayer.getLogicalDeviceId());

            APdu aarqAPdu = new InitialmessageProcessor(connectionData, dlmsLogicalDevice)
                    .processInitialMessage(payload);

            associationMessenger.encodeAndSend(aarqAPdu);

            Map<COSEMpdu.Choices, RequestProcessor> requestProcessors = buildRequestProcessors(associationMessenger,
                    requestProcessorData);

            while (true) {
                readNextMessage(associationMessenger, requestProcessors);
            }

        } catch (GenericAssociationException e) {
            // TODO remove statement
            System.out.println("**error in association**");
            try {
                associationMessenger.encodeAndSend(e.getErrorMessageApdu());
            } catch (IOException e1) {
                // ignore any exception here..
            }
        } catch (IOException e) {
            if (e instanceof EOFException) {
                System.out.println("client closed the connection..");
                return;
            }
            if (e instanceof SocketTimeoutException) {
                System.out.println("client was too slow.");
                return;
            }

            e.printStackTrace();
            // TODO ignore??
        } finally {
            this.semaphore.release();
            this.directory.removeConnection(this.connectionId);
            try {
                sessionLayer.close();
            } catch (IOException e1) {
                // ignore
            }

            notifyListener(connectionListener, Status.CLOSED);
        }
    }

    private void notifyListener(ServerConnectionListener connectionListener, Status status) {
        if (connectionListener == null) {
            return;
        }

        serverConnectionInformation.clientId = this.sessionLayer.getClientId();
        serverConnectionInformation.logicalDeviceAddress = this.sessionLayer.getLogicalDeviceId();
        serverConnectionInformation.status = status;
        connectionListener.connectionChanged(serverConnectionInformation);
    }

    private Map<COSEMpdu.Choices, RequestProcessor> buildRequestProcessors(AssociationMessenger associationMessenger,
            RequestProcessorData requestProcessorData) {
        Map<COSEMpdu.Choices, RequestProcessor> requestProcessors = new HashMap<>();
        requestProcessors.put(COSEMpdu.Choices.ACTION_REQUEST,
                new ActionRequestProcessor(associationMessenger, requestProcessorData));
        requestProcessors.put(COSEMpdu.Choices.GET_REQUEST,
                new GetRequestProcessor(associationMessenger, requestProcessorData));
        requestProcessors.put(COSEMpdu.Choices.SET_REQUEST,
                new SetRequestProcessor(associationMessenger, requestProcessorData));
        return requestProcessors;
    }

    private void readNextMessage(AssociationMessenger associationMessenger,
            Map<Choices, RequestProcessor> requestProcessors) throws IOException, AssociationException {

        COSEMpdu cosemPdu = associationMessenger.readNextCosemPdu();

        if (!this.connectionData.authenticated && cosemPdu.getChoiceIndex() != COSEMpdu.Choices.ACTION_REQUEST) {
            throw new AssociationException(StateError.SERVICE_NOT_ALLOWED, ServiceError.OPERATION_NOT_POSSIBLE);
        }

        RequestProcessor requestProcessor = requestProcessors.get(cosemPdu.getChoiceIndex());
        if (requestProcessor != null) {
            requestProcessor.processRequest(cosemPdu);
        }
        else {
            // TODO handle other requests..
        }

        if (!this.connectionData.authenticated) {
            throw new IOException("Client failed to authenticate..");
        }

    }

}
