package org.openmuc.jdlms.internal.lnassociation;

import java.io.IOException;
import java.util.Arrays;

import org.openmuc.jdlms.SecuritySuite;
import org.openmuc.jdlms.SecuritySuite.EncryptionMechanism;
import org.openmuc.jdlms.internal.APdu;
import org.openmuc.jdlms.internal.DataDirectory;
import org.openmuc.jdlms.internal.ServerConnectionData;
import org.openmuc.jdlms.internal.asn1.axdr.AxdrType;
import org.openmuc.jdlms.internal.asn1.axdr.NullOutputStream;
import org.openmuc.jdlms.internal.asn1.cosem.COSEMpdu;
import org.openmuc.jdlms.internal.sessionlayer.ServerSessionLayer;

public class AssociationMessenger {

    private final ServerConnectionData connectionData;
    private final DataDirectory directory;
    private final ServerSessionLayer sessionLayer;

    private SecuritySuite securitySuite;

    public AssociationMessenger(ServerConnectionData connectionData, DataDirectory directory,
            ServerSessionLayer sessionLayer, Long connectionId) {
        this.connectionData = connectionData;
        this.directory = directory;
        this.sessionLayer = sessionLayer;
    }

    public void encodeAndSend(APdu aPdu) throws IOException {
        send(encodeApdu(aPdu));
    }

    public void send(byte[] data) throws IOException {
        this.sessionLayer.send(data, 0, data.length);
    }

    public COSEMpdu readNextCosemPdu() throws IOException {
        byte[] bytes = this.sessionLayer.readNextMessage();
        APdu apdu = APdu.decode(bytes);

        SecuritySuite sec = connectionData.securitySuite;
        if (sec.getEncryptionMechanism() != EncryptionMechanism.NONE) {
            if (!aPduIsEncrypted(apdu)) {
                // TODO throw an ex
            }

            if (connectionData.clientSystemTitle == null) {
                connectionData.clientSystemTitle = systemTitle();
            }

            this.securitySuite = sec.clone();

            apdu = APdu.decode(bytes, connectionData.clientSystemTitle, connectionData.frameCounter, sec);
        }

        return apdu.cosemPdu;
    }

    private boolean aPduIsEncrypted(APdu apdu) {
        switch (apdu.cosemPdu.getChoiceIndex()) {
        case GLO_ACTION_REQUEST:
        case GLO_ACTION_RESPONSE:
        case GLO_EVENT_NOTIFICATION_REQUEST:
        case GLO_GET_REQUEST:
        case GLO_GET_RESPONSE:
        case GLO_INITIATEREQUEST:
        case GLO_INITIATERESPONSE:
        case GLO_READREQUEST:
        case GLO_READRESPONSE:
        case GLO_SET_REQUEST:
        case GLO_SET_RESPONSE:
        case GLO_WRITEREQUEST:
        case GLO_WRITERESPONSE:
        case DED_ACTIONREQUEST:
        case DED_ACTION_RESPONSE:
        case DED_EVENT_NOTIFICATION_REQUEST:
        case DED_GET_REQUEST:
        case DED_GET_RESPONSE:
        case DED_SET_REQUEST:
        case DED_SET_RESPONSE:
            return true;

        case ACTION_REQUEST:
        case ACTION_RESPONSE:
        case CONFIRMEDSERVICEERROR:
        case EVENT_NOTIFICATION_REQUEST:
        case EXCEPTION_RESPONSE:
        case GET_REQUEST:
        case GET_RESPONSE:
        case INFORMATIONREPORTREQUEST:
        case INITIATEREQUEST:
        case INITIATERESPONSE:
        case READREQUEST:
        case READRESPONSE:
        case SET_REQUEST:
        case SET_RESPONSE:
        case UNCONFIRMEDWRITEREQUEST:
        case WRITEREQUEST:
        case WRITERESPONSE:
        case _ERR_NONE_SELECTED:
        default:
            return false;
        }
    }

    public byte[] encodeApdu(APdu aPdu) throws IOException {
        // TODO: better user of the buffer
        byte[] buffer = new byte[0xFFFFF];
        int length = 0;

        SecuritySuite sec = connectionData.securitySuite;
        if (sec.getEncryptionMechanism() != EncryptionMechanism.NONE) {

            if (this.securitySuite == null) {
                this.securitySuite = sec.clone();
            }

            length = aPdu.encode(buffer, connectionData.frameCounter++, systemTitle(), this.securitySuite);
        }
        else {
            length = aPdu.encode(buffer);
        }

        return Arrays.copyOfRange(buffer, buffer.length - length, buffer.length);
    }

    public byte[] systemTitle() {
        return this.directory.getLogicalDeviceFor(this.sessionLayer.getLogicalDeviceId())
                .getLogicalDevice()
                .getSystemTitle();
    }

    public boolean pduSizeTooLarge(AxdrType actionResponse) throws IOException {
        return this.connectionData.clientMaxReceivePduSize != 0
                && pduSizeOf(actionResponse) >= this.connectionData.clientMaxReceivePduSize;
    }

    public static int pduSizeOf(AxdrType pdu) throws IOException {
        return pdu.encode(new NullOutputStream());
    }

}
