package org.openmuc.jdlms.internal.lnassociation;

import org.openmuc.jasn1.ber.types.BerAnyNoDecode;
import org.openmuc.jasn1.ber.types.BerBitString;
import org.openmuc.jasn1.ber.types.BerInteger;
import org.openmuc.jasn1.ber.types.BerOctetString;
import org.openmuc.jdlms.internal.APdu;
import org.openmuc.jdlms.internal.AssociateSourceDiagnostic.AcseServiceProvider;
import org.openmuc.jdlms.internal.AssociateSourceDiagnostic.AcseServiceUser;
import org.openmuc.jdlms.internal.AssociationResult;
import org.openmuc.jdlms.internal.ContextId;
import org.openmuc.jdlms.internal.ObjectIdentifier;
import org.openmuc.jdlms.internal.asn1.cosem.COSEMpdu;
import org.openmuc.jdlms.internal.asn1.cosem.Conformance;
import org.openmuc.jdlms.internal.asn1.cosem.InitiateResponse;
import org.openmuc.jdlms.internal.asn1.cosem.Integer16;
import org.openmuc.jdlms.internal.asn1.cosem.Unsigned16;
import org.openmuc.jdlms.internal.asn1.cosem.Unsigned8;
import org.openmuc.jdlms.internal.asn1.iso.acse.AARE_apdu;
import org.openmuc.jdlms.internal.asn1.iso.acse.ACSE_apdu;
import org.openmuc.jdlms.internal.asn1.iso.acse.AP_title;
import org.openmuc.jdlms.internal.asn1.iso.acse.AP_title_form2;
import org.openmuc.jdlms.internal.asn1.iso.acse.Associate_result;
import org.openmuc.jdlms.internal.asn1.iso.acse.Associate_source_diagnostic;
import org.openmuc.jdlms.internal.asn1.iso.acse.Association_information;
import org.openmuc.jdlms.internal.asn1.iso.acse.Authentication_value;

class InitialResponseBuilder {

    private Conformance conformanceSetting;
    private ContextId contextId;
    private Associate_result result;
    private Associate_source_diagnostic associateSourceDiagnostic;
    private Association_information userInformation;
    private AP_title respondingAPTitle;
    private Authentication_value serverToClientChallenge;

    public InitialResponseBuilder(Conformance conformanceSetting) {
        this.conformanceSetting = conformanceSetting;

        this.contextId = ContextId.LOGICAL_NAME_REFERENCING_WITH_CIPHERING;

        this.result = new Associate_result(AssociationResult.ACCEPTED.getValue());

        this.associateSourceDiagnostic = new Associate_source_diagnostic(new BerInteger(AcseServiceUser.NULL.getCode()),
                null);

        this.userInformation = new Association_information(new BerAnyNoDecode());

        this.respondingAPTitle = null;
    }

    public InitialResponseBuilder setConformanceSetting(Conformance conformanceSetting) {
        this.conformanceSetting = conformanceSetting;
        return this;
    }

    public InitialResponseBuilder setResult(AssociationResult result) {
        this.result = new Associate_result(result.getValue());
        return this;
    }

    public InitialResponseBuilder setSystemTitle(byte[] respondingAPTitle) {
        this.respondingAPTitle = new AP_title(new AP_title_form2(respondingAPTitle));
        return this;
    }

    public InitialResponseBuilder setAuthenticationValue(byte[] serverToClientChallenge) {
        this.serverToClientChallenge = new Authentication_value(new BerOctetString(serverToClientChallenge), null);
        return this;
    }

    public InitialResponseBuilder setContextId(ContextId contextId) {
        this.contextId = contextId;
        return this;
    }

    public InitialResponseBuilder setAssociateSourceDiagnostic(AcseServiceUser acseServiceUser) {
        return setAssociateSourceDiagnostic(acseServiceUser, null);
    }

    public InitialResponseBuilder setAssociateSourceDiagnostic(AcseServiceProvider acseServiceProvider) {
        return setAssociateSourceDiagnostic(null, acseServiceProvider);
    }

    private InitialResponseBuilder setAssociateSourceDiagnostic(AcseServiceUser acseServiceUser,
            AcseServiceProvider acseServiceProvider) {
        BerInteger serviceUser = null;
        BerInteger serviceProvider = null;
        if (acseServiceUser != null) {
            serviceUser = new BerInteger(acseServiceUser.getCode());
        }
        if (acseServiceProvider != null) {
            serviceProvider = new BerInteger(acseServiceProvider.getCode());
        }

        this.associateSourceDiagnostic = new Associate_source_diagnostic(serviceUser, serviceProvider);
        return this;
    }

    public InitialResponseBuilder setUserInformation(Association_information userInformation) {
        this.userInformation = userInformation;
        return this;
    }

    public APdu build() {

        COSEMpdu xDlmsInitiateResponsePdu = new COSEMpdu();

        InitiateResponse initiateResponse = new InitiateResponse(null, new Unsigned8(6), this.conformanceSetting,
                new Unsigned16(0xffff), new Integer16(0x0007));
        xDlmsInitiateResponsePdu.setinitiateResponse(initiateResponse);

        AARE_apdu aare = new AARE_apdu();
        aare.protocol_version = new BerBitString(new byte[] { (byte) 0x80 }, 1);
        aare.application_context_name = ObjectIdentifier.applicationContextNameFor(this.contextId);

        aare.result = this.result;

        aare.result_source_diagnostic = this.associateSourceDiagnostic;

        aare.user_information = this.userInformation;

        aare.responding_AP_title = this.respondingAPTitle;

        aare.responding_authentication_value = this.serverToClientChallenge;

        ACSE_apdu aarqAcseAPdu = new ACSE_apdu(null, aare, null, null);

        return new APdu(aarqAcseAPdu, xDlmsInitiateResponsePdu);
    }

}
