package org.openmuc.jdlms.internal.transportlayer;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Semaphore;
import java.util.concurrent.ThreadPoolExecutor;

import org.openmuc.jdlms.internal.DataDirectory;
import org.openmuc.jdlms.internal.lnassociation.Association;
import org.openmuc.jdlms.internal.sessionlayer.ServerSessionLayer;
import org.openmuc.jdlms.internal.sessionlayer.ServerWrapperLayer;
import org.openmuc.jdlms.internal.settings.TcpServerSettings;
import org.openmuc.jdlms.transportlayer.client.StreamAccessor;

public class ServerTcpLayer implements ServerTransportLayer {

    private SocketListener socketListener;
    private final DataDirectory directory;
    private ExecutorService singleThreadExecutor;
    private final TcpServerSettings settings;

    public ServerTcpLayer(DataDirectory directory, TcpServerSettings settings) {
        this.directory = directory;
        this.settings = settings;
    }

    @Override
    public void close() throws IOException {
        this.socketListener.close();
        this.singleThreadExecutor.shutdown();
    }

    @Override
    public void start() throws IOException {
        this.socketListener = new SocketListener();
        singleThreadExecutor = Executors.newSingleThreadExecutor();
        singleThreadExecutor.submit(socketListener);
    }

    public static class TcpServerConnectionInformation extends ServerConnectionInformationImpl {

        public InetAddress clienInetAddress;

        public TcpServerConnectionInformation(InetAddress clienInetAddress) {
            this.clienInetAddress = clienInetAddress;
        }

        @Override
        public InetAddress getClienInetAddress() {
            return this.clienInetAddress;
        }
    };

    private class SocketListener implements Runnable, AutoCloseable {

        private final ServerSocket serverSocket;
        private final ThreadPoolExecutor threadPool;
        private final Semaphore semaphore;

        private long connections;

        public SocketListener() throws IOException {
            this.serverSocket = new ServerSocket(settings.tcpPort);
            this.serverSocket.setSoTimeout(0);

            int permits = settings.maxClients == 0 ? Integer.MAX_VALUE : settings.maxClients;
            this.threadPool = (ThreadPoolExecutor) Executors.newFixedThreadPool(permits);
            this.semaphore = new Semaphore(permits, true);
            this.connections = 0L;
        }

        @Override
        public void run() {
            try {
                while (true) {
                    // the semaphore is released, when the client closes
                    this.semaphore.acquire();
                    Socket socket = serverSocket.accept();

                    ServerSessionLayer sessionLayer = new ServerWrapperLayer(new TcpSocketAccessor(socket), settings);

                    Long connectionId = ++connections;

                    Association association = new Association(directory, sessionLayer, this.semaphore, connectionId,
                            settings, new TcpServerConnectionInformation(socket.getInetAddress()));
                    this.threadPool.execute(association);
                }
            } catch (IOException e) {
                // ignore here, connection will be closed
            } catch (InterruptedException e) {
                // shouldn't occur.
            }
        }

        @Override
        public void close() throws IOException {
            this.serverSocket.close();
            this.threadPool.shutdown();
        }
    }

    private class TcpSocketAccessor implements StreamAccessor {

        private final Socket socket;

        public TcpSocketAccessor(Socket socket) {
            this.socket = socket;
        }

        @Override
        public void setTimeout(int timeout) throws IOException {
            this.socket.setSoTimeout(timeout);
        }

        @Override
        public InputStream getInputStream() throws IOException {
            return this.socket.getInputStream();
        }

        @Override
        public OutputStream getOutpuStream() throws IOException {
            return this.socket.getOutputStream();
        }

        @Override
        public void close() throws IOException {
            this.socket.close();
        }

    }
}
