/*
 * Copyright 2012-16 Fraunhofer ISE
 *
 * This file is part of jDLMS.
 * For more information visit http://www.openmuc.org
 *
 * jDLMS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * jDLMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with jDLMS.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.jdlms.internal;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.MessageFormat;

import org.openmuc.jasn1.ber.BerByteArrayOutputStream;
import org.openmuc.jasn1.ber.types.BerAnyNoDecode;
import org.openmuc.jdlms.FatalJDlmsException;
import org.openmuc.jdlms.JDlmsException.ExceptionId;
import org.openmuc.jdlms.JDlmsException.Fault;
import org.openmuc.jdlms.SecuritySuite;
import org.openmuc.jdlms.internal.AssociateSourceDiagnostic.AcseServiceProvider;
import org.openmuc.jdlms.internal.AssociateSourceDiagnostic.AcseServiceUser;
import org.openmuc.jdlms.internal.asn1.axdr.AxdrLength;
import org.openmuc.jdlms.internal.asn1.cosem.COSEMpdu;
import org.openmuc.jdlms.internal.asn1.iso.acse.ACSE_apdu;
import org.openmuc.jdlms.internal.asn1.iso.acse.Associate_source_diagnostic;
import org.openmuc.jdlms.internal.asn1.iso.acse.Association_information;
import org.openmuc.jdlms.internal.lnassociation.AssociatRequestException;
import org.openmuc.jdlms.internal.security.GcmModule;

public class APdu {
    private static final SecuritySuite DEFAULT_SECURITY_SUITE = SecuritySuite.builder().build();

    public ACSE_apdu acseAPdu;
    public COSEMpdu cosemPdu;

    private APdu() {
    }

    public static APdu decode(byte[] bytes, byte[] serverSystemTitle, int frameCounter, SecuritySuite securitySuite)
            throws IOException {
        return decode(bytes, true, serverSystemTitle, frameCounter, securitySuite);
    }

    public static APdu decode(byte[] bytes) throws IOException {
        return decode(bytes, false, null, 0, DEFAULT_SECURITY_SUITE);
    }

    private static APdu decode(byte[] bytes, boolean encrypt, byte[] serverSystemTitle, int frameCounter,
            SecuritySuite securitySuite) throws IOException {
        DataInputStream is = new DataInputStream(new ByteArrayInputStream(bytes));

        APdu aPdu = new APdu();

        int tag = bytes[0] & 0xFF;
        if (tag >= 0x60 && tag <= 0x63) {
            aPdu.acseAPdu = new ACSE_apdu();
            aPdu.acseAPdu.decode(is, null);

            validateAssociateResult(aPdu);

            if (encrypt) {
                if (aPdu.acseAPdu.aare != null && aPdu.acseAPdu.aare.responding_AP_title != null) {
                    serverSystemTitle = aPdu.acseAPdu.aare.responding_AP_title.ap_title_form2.value;
                }
                else if (aPdu.acseAPdu.aarq != null) {
                    ContextId contextId = ObjectIdentifier
                            .applicationContextIdFor(aPdu.acseAPdu.aarq.application_context_name);

                    if (!contextId.isCiphered()) {
                        throw new AssociatRequestException(AcseServiceUser.APPLICATION_CONTEXT_NAME_NOT_SUPPORTED);
                    }

                    if (aPdu.acseAPdu.aarq.calling_AP_title != null) {
                        serverSystemTitle = aPdu.acseAPdu.aarq.calling_AP_title.ap_title_form2.value;

                    }
                }
            }
        }

        InputStream cosemPduIs = is;
        if (encrypt) {

            is.read();

            AxdrLength axdrLength = new AxdrLength();
            axdrLength.decode(is);
            int encLength = axdrLength.getValue();

            byte[] ciphertext = new byte[encLength];
            is.readFully(ciphertext);

            byte[] plaintext = GcmModule.decrypt(ciphertext, serverSystemTitle, securitySuite);

            cosemPduIs = new ByteArrayInputStream(plaintext);
        }

        if (cosemPduIs.available() > 0) {
            aPdu.cosemPdu = new COSEMpdu();
            aPdu.cosemPdu.decode(cosemPduIs);
        }

        return aPdu;
    }

    private static void validateAssociateResult(APdu aPdu) throws FatalJDlmsException {
        ACSE_apdu acseAPdu = aPdu.acseAPdu;
        if (acseAPdu.aare == null) {
            return;
        }

        AssociationResult associationResult = AssociationResult.associationResultFor(acseAPdu.aare.result.value);

        if (associationResult == AssociationResult.ACCEPTED) {
            return;
        }

        Associate_source_diagnostic resultSourceDiagnostic = acseAPdu.aare.result_source_diagnostic;
        System.out.println(aPdu.acseAPdu.aare);
        Fault faut;
        ExceptionId exceptionId;
        System.out.println(resultSourceDiagnostic);
        if (resultSourceDiagnostic.acse_service_user != null) {
            faut = Fault.USER;
            AcseServiceUser serviceUser = AssociateSourceDiagnostic.AcseServiceUser
                    .acseServiceUserFor(resultSourceDiagnostic.acse_service_user.value);

            switch (serviceUser) {
            case APPLICATION_CONTEXT_NAME_NOT_SUPPORTED:
                exceptionId = ExceptionId.WRONG_REFERENCING_METHOD;
                break;

            case AUTHENTICATION_FAILURE:
            case AUTHENTICATION_MECHANISM_NAME_NOT_RECOGNISED:
            case AUTHENTICATION_MECHANISM_NAME_REQUIRED:
                exceptionId = ExceptionId.AUTHENTICATION_ERROR;
                break;// TODO: genauer

            case AUTHENTICATION_REQUIRED:
                exceptionId = ExceptionId.AUTHENTICATION_REQUIRED;
                break;

            default:
            case NO_REASON_GIVEN:
            case NULL:
                exceptionId = ExceptionId.UNKNOWN;
                break;
            }
        }
        else {
            faut = Fault.SYSTEM;
            AcseServiceProvider acseServiceProvider = AcseServiceProvider
                    .acseServiceProviderFor(resultSourceDiagnostic.acse_service_provider.value);

            switch (acseServiceProvider) {
            case NO_COMMON_ACSE_VERSION:
            case NO_REASON_GIVEN:
            case NULL:
            default:
                exceptionId = ExceptionId.UNKNOWN;
                break;

            // TODO: genauer
            }
        }
        String message = MessageFormat.format(
                "Received an association response (AARE) with an error message. Result name {0}.", associationResult);
        throw new FatalJDlmsException(exceptionId, faut, message);
    }

    public APdu(ACSE_apdu acseAPdu, COSEMpdu cosemPdu) {
        this.acseAPdu = acseAPdu;
        this.cosemPdu = cosemPdu;
    }

    public int encode(byte[] buffer, int frameCounter, byte[] systemTitle, SecuritySuite securitySuite)
            throws IOException {
        BerByteArrayOutputStream baos = new BerByteArrayOutputStream(buffer, buffer.length - 1);
        int numBytesEncoded = encodeCosemPdu(buffer, baos);

        // -- encrypting
        int origTag = buffer[buffer.length - numBytesEncoded] & 0xff;
        int newTag;

        if (origTag < 25) {
            newTag = origTag + 32;
        }
        else {
            newTag = origTag + 8;
        }
        byte[] ciphertext = GcmModule.processPlain(buffer, buffer.length - numBytesEncoded, numBytesEncoded, systemTitle,
                frameCounter, securitySuite, (byte) newTag);

        numBytesEncoded = ciphertext.length;

        System.arraycopy(ciphertext, 0, buffer, buffer.length - numBytesEncoded, ciphertext.length);
        baos = new BerByteArrayOutputStream(buffer, buffer.length - numBytesEncoded - 1);
        // -- encrypting

        return encodeAcsePdu(numBytesEncoded, baos);
    }

    public int encode(byte[] buffer) throws IOException {
        BerByteArrayOutputStream baos = new BerByteArrayOutputStream(buffer, buffer.length - 1);
        int numBytesEncoded = encodeCosemPdu(buffer, baos);
        return encodeAcsePdu(numBytesEncoded, baos);
    }

    private int encodeCosemPdu(byte[] buffer, BerByteArrayOutputStream baos) throws IOException {
        return cosemPdu.encode(baos);
    }

    private int encodeAcsePdu(int numBytesEncoded, BerByteArrayOutputStream baos) throws IOException {
        if (acseAPdu != null) {
            if (acseAPdu.aarq != null) {
                acseAPdu.aarq.user_information = new Association_information(new BerAnyNoDecode(numBytesEncoded));
            }
            else if (acseAPdu.aare != null) {
                acseAPdu.aare.user_information = new Association_information(new BerAnyNoDecode(numBytesEncoded));
            }
            numBytesEncoded = acseAPdu.encode(baos, true);
        }
        return numBytesEncoded;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        if (acseAPdu != null) {
            sb.append("ACSE PDU: ").append(acseAPdu.toString()).append(", ");
        }
        return sb.append("COSEM xDLMS PDU:").append(cosemPdu.toString()).toString();
    }

}
