package org.openmuc.jdlms.sessionlayer;

import static java.lang.String.format;
import static org.openmuc.jdlms.JDlmsException.Fault.SYSTEM;

import java.io.DataInputStream;
import java.io.EOFException;
import java.io.IOException;
import java.net.SocketException;
import java.nio.ByteBuffer;

import org.openmuc.jdlms.FatalJDlmsException;
import org.openmuc.jdlms.JDlmsException.ExceptionId;
import org.openmuc.jdlms.internal.sessionlayer.SessionLayerListener;
import org.openmuc.jdlms.sessionlayer.WrapperHeader.WrapperHeaderBuilder;
import org.openmuc.jdlms.settings.client.Settings;
import org.openmuc.jdlms.transportlayer.client.TransportLayer;

public class WrapperLayer implements SessionLayer {

    private final WrapperHeaderBuilder headerBuilder;
    private final TransportLayer transportLayer;

    private SessionLayerListener eventListener;
    private final Settings settings;
    private final boolean closed;

    public WrapperLayer(Settings settings, TransportLayer transportLayer) throws IOException {
        this.settings = settings;
        this.headerBuilder = WrapperHeader.builder(settings.clientId(), settings.logicalDeviceId());
        this.transportLayer = transportLayer;

        this.closed = true;
    }

    @Override
    public void startListening(SessionLayerListener eventListener) throws IOException {
        this.eventListener = eventListener;
        if (!closed) {
            return;
        }

        transportLayer.open();

        Thread readerThread = new Thread(new ConnectionReader());
        readerThread.setName("WRAPPER/TCP-CONNECTION_READER");
        readerThread.start();
    }

    @Override
    public void send(byte[] tSdu, int off, int len) throws IOException {
        this.headerBuilder.setLength(len);

        byte[] wpdu = ByteBuffer.allocate(len + WrapperHeader.HEADER_LENGTH)
                .put(headerBuilder.build().encode())
                .put(tSdu, off, len)
                .array();
        this.transportLayer.getOutpuStream().write(wpdu);
        this.transportLayer.getOutpuStream().flush();
    }

    @Override
    public void close() throws IOException {
        this.transportLayer.close();
    }

    private void validate(WrapperHeader header, int payloadLength) throws FatalJDlmsException {
        if (header.sourceWPort() != settings.logicalDeviceId()) {
            throw new FatalJDlmsException(ExceptionId.WRAPPER_HEADER_INVALID_SRC_DEST_ADDR, SYSTEM,
                    String.format(
                            "Connection was initiated with logical device address %d, but server answered with %d",
                            settings.logicalDeviceId(), header.sourceWPort()));
        }

        if (payloadLength != header.payloadLength()) {
            throw new FatalJDlmsException(ExceptionId.WRAPPER_HEADER_INVALID_PAYLOAD_LENGTH, SYSTEM,
                    format("Header specified length to be %d, but was %d.", header.payloadLength(), payloadLength));
        }

    }

    private class ConnectionReader implements Runnable {

        @Override
        public void run() {
            try {
                mainLoop();
            } catch (EOFException e) {
                EOFException closedIOException;
                if (!closed) {
                    closedIOException = new EOFException("Socket was closed by remote host.");
                    eventListener.connectionInterrupted(closedIOException);
                }
            } catch (IOException e) {
                eventListener.connectionInterrupted(e);
            } finally {
                unsaveClose();
            }
        }

        private void mainLoop() throws SocketException, IOException {
            DataInputStream inputStream = new DataInputStream(transportLayer.getInputStream());
            while (true) {
                if (transportLayer.isClosed()) {
                    return;
                }

                transportLayer.setTimeout(0);
                WrapperHeader header = WrapperHeader.decode(transportLayer, settings.responseTimeout());
                byte[] messageData = new byte[header.payloadLength()];
                inputStream.readFully(messageData);
                validate(header, messageData.length);

                eventListener.dataReceived(messageData);

                Thread.yield();
            }
        }

    }

    private void unsaveClose() {
        if (!closed) {
            try {
                close();
            } catch (IOException e) {
                // ignore
            }
        }
    }

}
